from .errors import NotSupportedError
from .impl_registry import get_implementation

__all__ = ("adrenaline", "is_sleep_prevented", "prevent_sleep", "NotSupportedError")

_impl = get_implementation()

adrenaline = _impl.prevent_sleep
prevent_sleep = _impl.prevent_sleep
is_sleep_prevented = _impl.is_sleep_prevented



__all__ = ("_enter", "_exit", "_verify")

from contextlib import closing
from typing import Any, Callable, List

from jeepney import DBusErrorResponse, MessageGenerator, new_method_call
from jeepney.io.blocking import Proxy, open_dbus_connection

from ..errors import NotSupportedError

#: Type specification for power management cookies
PowerManagementCookie = int


class FreedesktopPowerManagement(MessageGenerator):
    interface = "org.freedesktop.PowerManagement"

    def __init__(
        self,
        object_path: str = "/org/freedesktop/PowerManagement",
        bus_name: str = "org.freedesktop.PowerManagement",
    ):
        super().__init__(object_path, bus_name)

    def inhibit(self, app_name: str, reason: str) -> PowerManagementCookie:
        return new_method_call(self, "Inhibit", "ss", (app_name, reason))

    def uninhibit(self, cookie: PowerManagementCookie) -> None:
        return new_method_call(self, "Uninhibit", "u", (cookie,))


class GNOMESessionManager(MessageGenerator):
    interface = "org.gnome.SessionManager"

    TOPLEVEL_XID = 0
    INHIBIT_SUSPEND = 4

    def __init__(
        self,
        object_path: str = "/org/gnome/SessionManager",
        bus_name: str = "org.gnome.SessionManager",
    ):
        super().__init__(object_path, bus_name)

    def inhibit(self, app_name: str, reason: str) -> PowerManagementCookie:
        return new_method_call(
            self,
            "Inhibit",
            "susu",
            (
                app_name,
                GNOMESessionManager.TOPLEVEL_XID,
                reason,
                GNOMESessionManager.INHIBIT_SUSPEND,
            ),
        )

    def uninhibit(self, cookie: PowerManagementCookie) -> None:
        return new_method_call(self, "Uninhibit", "u", (cookie,))

    def is_inhibited(self) -> bool:
        return new_method_call(
            self, "IsInhibited", "u", (GNOMESessionManager.INHIBIT_SUSPEND,)
        )


_connection = None
_disposers: List[Callable[[Any], None]] = []
_interfaces = []

_interface_candidates = [GNOMESessionManager, FreedesktopPowerManagement]


def _enter(*, display: bool, app_name: str, reason: str) -> None:
    global _connection, _disposers, _interfaces, _interface_candidates

    if _connection is None:
        try:
            _connection = open_dbus_connection()
        except Exception:
            raise NotSupportedError("Cannot establish connection to D-Bus") from None

    success = False
    try:
        for cls in _interface_candidates:
            interface = cls()
            try:
                (cookie,) = Proxy(interface, _connection).inhibit(app_name, reason)
                success = True
                break
            except DBusErrorResponse:
                pass
        else:
            raise NotSupportedError(
                "No supported power management DBus interface is available"
            )
    finally:
        if not success and not _disposers and _connection is not None:
            # connection is not needed any more
            _connection.close()
            _connection = None

    def disposer(bus: Any) -> None:
        Proxy(interface, bus).uninhibit(cookie)

    _disposers.append(disposer)
    _interfaces.append(interface)


def _exit() -> None:
    global _connection, _disposers

    assert _connection

    _interfaces.pop()
    disposer = _disposers.pop()
    disposer(_connection)

    if not _disposers:
        # connection is not needed any more
        _connection.close()
        _connection = None


def _verify() -> bool:
    global _interfaces

    if not _interfaces:
        return False

    interface = _interfaces[-1]
    if hasattr(interface, "is_inhibited"):
        with closing(open_dbus_connection()) as bus:
            try:
                (result,) = Proxy(interface, bus).is_inhibited()
                return result
            except DBusErrorResponse:
                pass

    global _disposers
    return bool(_disposers)



import ctypes

__all__ = ("_enter", "_exit", "_verify")

# The following constants are from
# https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-setthreadexecutionstate

ES_CONTINUOUS = 0x80000000
ES_SYSTEM_REQUIRED = 0x00000001
ES_DISPLAY_REQUIRED = 0x00000002

_previous_state = None


def _enter(*, display: bool, app_name: str, reason: str) -> None:
    global _previous_state

    flags = ES_CONTINUOUS | ES_SYSTEM_REQUIRED
    if display:
        flags |= ES_DISPLAY_REQUIRED

    previous_state = ctypes.windll.kernel32.SetThreadExecutionState(flags)
    if previous_state is None:
        raise RuntimeError("Failed to call SetThreadExecutionState()")

    _previous_state = previous_state


def _exit() -> None:
    assert _previous_state is not None
    ctypes.windll.kernel32.SetThreadExecutionState(_previous_state)


def _verify() -> bool:
    previous_state = ctypes.windll.kernel32.SetThreadExecutionState(0)
    if previous_state is None:
        raise RuntimeError("Failed to call SetThreadExecutionState()")
    ctypes.windll.kernel32.SetThreadExecutionState(previous_state)
    return bool(previous_state & (ES_SYSTEM_REQUIRED | ES_DISPLAY_REQUIRED))



#!/usr/bin/env python3
"""Command line tool that prevents the computer from sleeping while the
script is running.
"""

import platform
import sys
from argparse import ArgumentParser

from adrenaline import adrenaline


def decorate(text: str):
    # Linux and macOS terminals support emojis
    decoration = "\u2728" if platform.system() != "Windows" else ">"
    lines = text.strip().split("\n")
    for index, line in enumerate(lines):
        print(f"   {line}" if index else f"{decoration} {line}")


def main():
    parser = ArgumentParser(description=sys.modules[__name__].__doc__)

    parser.add_argument(
        "--display",
        action="store_true",
        default=False,
        help="forces the display to stay on as well",
    )

    options = parser.parse_args()

    with adrenaline(display=options.display):
        message = "Sleep mode inhibited."
        if options.display:
            message += " The display is also forced to stay on."

        decorate(f"{message}\nPress Enter to quit.")
        try:
            input()
        except KeyboardInterrupt:
            pass


if __name__ == "__main__":
    main()



"""Implementation of the platform-dependent part of this module for Darwin."""

# Based on the work of Michael Lynn
# https://github.com/pudquick/pypmset/blob/master/pypmset.py

from ctypes import POINTER, byref, c_uint32, c_void_p, cdll
from ctypes.util import find_library

from CoreFoundation import (  # type: ignore
    CFStringCreateWithCString,
    kCFStringEncodingASCII,
)
from objc import pyobjc_id  # type: ignore

__all__ = ("_enter", "_exit", "_verify")


# Load the IOKit library
libIOKit = cdll.LoadLibrary(find_library("IOKit"))

# Define argument types for the IOKit functions we are going to use
libIOKit.IOPMAssertionCreateWithName.argtypes = [
    c_void_p,
    c_uint32,
    c_void_p,
    POINTER(c_uint32),
]
libIOKit.IOPMCopyAssertionsStatus.argtypes = [c_void_p]
libIOKit.IOPMAssertionRelease.argtypes = [c_uint32]


def CFSTR(py_string: bytes):
    """Converts a Python byte string into an Objective C string, using ASCII
    encoding.
    """
    return CFStringCreateWithCString(None, py_string, kCFStringEncodingASCII)


def raw_ptr_from_string(pyobjc_string):
    return pyobjc_id(pyobjc_string.nsstring())


def IOPMAssertionCreateWithName(type: bytes, level: int, name: bytes) -> c_uint32:
    assertion_id = c_uint32(0)
    p_assertion_type = raw_ptr_from_string(CFSTR(type))
    p_assertion_name = raw_ptr_from_string(CFSTR(name))
    error_code = libIOKit.IOPMAssertionCreateWithName(
        p_assertion_type, level, p_assertion_name, byref(assertion_id)
    )
    if error_code != 0:
        raise RuntimeError("Failed to create power management assertion")
    return assertion_id


def IOPMAssertionRelease(assertion_id: c_uint32) -> None:
    error_code = libIOKit.IOPMAssertionRelease(assertion_id)
    if error_code != 0:
        raise RuntimeError("Failed to release power management assertion")


def get_assertion_type(display: bool) -> bytes:
    """Returns the macOS power management assertion type suitable for the given
    configuration.

    Parameters:
        display: whether to keep the display on

    Returns:
        the macOS power management assertion type as a Python byte string
    """
    if display:
        return b"NoDisplaySleepAssertion"
    else:
        return b"NoIdleSleepAssertion"


kIOPMAssertionLevelOn = 255
current_assertion_id = c_uint32(0)


def _enter(*, display: bool, app_name: str, reason: str) -> None:
    global current_assertion_id
    assertion_type = get_assertion_type(display)
    if current_assertion_id.value == 0:
        current_assertion_id = IOPMAssertionCreateWithName(
            assertion_type,
            kIOPMAssertionLevelOn,
            reason.encode("ascii", errors="replace"),
        )


def _exit() -> None:
    global current_assertion_id
    IOPMAssertionRelease(current_assertion_id)
    current_assertion_id.value = 0


def _verify() -> bool:
    from subprocess import check_output

    output = check_output(["pmset", "-g", "assertions"])
    lines = output.strip().split(b"\n")
    return any(
        b'SleepAssertion named: "Python adrenaline module"' in line for line in lines
    )



__all__ = ("_enter", "_exit", "_verify")


def _enter(*args, **kwds) -> None:
    raise NotImplementedError("Not supported on this platform")


def _verify() -> bool:
    return False


_exit = _enter



__all__ = ("_enter", "_exit", "_verify")


_counter = 0


def _enter(*args, **kwds) -> None:
    global _counter
    _counter += 1


def _exit() -> None:
    global _counter
    assert _counter > 0
    _counter -= 1


def _verify() -> bool:
    global _counter
    return _counter > 0



import platform
from contextlib import contextmanager
from typing import Iterator, Optional

from .types import Adrenaline

__all__ = ("get_implementation",)


_implementations = {}  # type: Dict[str, Adrenaline]


def get_implementation(name: Optional[str] = None) -> Adrenaline:
    if name is None:
        name = platform.system()

    name = name.lower()
    impl = _implementations.get(name)
    if impl is None:
        _implementations[name] = impl = import_implementation(name)

    return impl


def import_implementation(name: str) -> Adrenaline:
    if name == "darwin":
        from ._impl.darwin import _enter, _exit, _verify
    elif name == "windows":
        from ._impl.windows import _enter, _exit, _verify
    elif name == "linux":
        from ._impl.linux import _enter, _exit, _verify
    elif name == "dummy":
        from ._impl.dummy import _enter, _exit, _verify
    elif name == "failing":
        from ._impl.failing import _enter, _exit, _verify
    else:
        raise NotImplementedError(f"No such implementation: {name!r}")

    class _AdrenalineImpl(Adrenaline):
        @contextmanager
        def prevent_sleep(
            self,
            *,
            display: bool = False,
            app_name: Optional[str] = None,
            reason: Optional[str] = None,
        ) -> Iterator[None]:
            _enter(
                display=display,
                app_name=app_name or "Python",
                reason=reason or "Python adrenaline module",
            )
            try:
                yield
            finally:
                _exit()

        def is_sleep_prevented(self) -> bool:
            return _verify()

    return _AdrenalineImpl()



from typing import ContextManager, Optional

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol


__all__ = ("Adrenaline", "AdrenalineContext")


#: Type specification for a context that suppresses sleep when entered and restores
#: the previous state when exited
AdrenalineContext = ContextManager[None]


class Adrenaline(Protocol):
    def is_sleep_prevented(self) -> bool:
        pass

    def prevent_sleep(
        self,
        *,
        display: bool = False,
        app_name: Optional[str] = None,
        reason: Optional[str] = None
    ) -> AdrenalineContext:
        pass



__all__ = ("NotSupportedError",)


class NotSupportedError(RuntimeError):
    """Error thrown when the power management API is not supported or implemented
    in the OS that the system is running on.
    """

    pass
